/* IZ 050626
utility for making a node (synth or group) perform some action when 
it is freed (removed) from the server. 

a = { SinOsc.ar(400, 0, 0.1) }.play;
a.onEnd { | who | Post << who << " ended!" << "\r"; };
// ------------
a.free;

a = Synth("variable_sin")
	.onEnd({ 'hi there i just stopped'.postln; })
	.onEnd({ { 'and there is even more stuff'.postln } ! 5});
a.free;

*/

// TODO: See if quit methods are still needed by Lilt (it seems not so, so they should be removed).
// Note: onEnd should definitely be kept.

+ Object {
	quit {} // objects ignore quit for general compatibility with Script
}

+ Node {
	// cannot use free because it requires boolean first arg.
	quit {
		this.free;
//		argFunc.(this); // rely on dependant for notification after end
	}
	onEnd { | argFunc |
		var cmdPeriodFunc;
		NodeWatcher.register(this);
		// this is needed to add this to catch synths stopped by Command-Period:
		cmdPeriodFunc = {
			NodeWatcher.unregister(this);
			CmdPeriod.remove(cmdPeriodFunc);
			argFunc.value;
		};
		CmdPeriod.add(cmdPeriodFunc);
		// This evaluates argFunc when notified by NodeWatcher
		this.addDependant { | me, whatHappened |
			if (whatHappened == \n_end) {
				argFunc.(this);
				this.removeDependant(argFunc);
				// see CmdPeriod above
				CmdPeriod.remove(cmdPeriodFunc);
			}
		};
	}
}

+ Stream {
	quit { | argFunc |
		this.stop;
		argFunc.(this); // other than nodes, must notify immediately
	}
}

+ EventStreamPlayer {
	onEnd { | argFunc |
		this.addDependant { | me, whatHappened |
			if (whatHappened == \stopped) {
				argFunc.(this);
				this.removeDependant(argFunc);
			}
		}
	}
}

+ Object { 
	/* Everything else should ignore onEnd */
	onEnd {} // ignore on end
}
