/* (IZ 2005-10-19) {
3d redo of list management Classes. Just the functionality for adding, removing, selecting, displaying, loading and saving data.

ListModel: list management functionality
ListWithGui: ListModel with additional name, window and window making function. 

======================================================================
not included because this part should be done in other class / code (!?)
// actions to do on this list when a keydown is received in a list view:
var >space, >return, >backspace, >enter, >escape;

} */

ListModel : Model {
	var <list;			// The actual collection of items to be managed
	// functions for converting data items etc. 
	var >add, >remove;	// functions for adding or removing items, depending on class of list
	var >makeNames;	// make name list for namesItems for display on list views
	var >fromData; 	// create new list from data (loaded from file)
	var >toData;		// convert list to data for saving on file
	// internal caches for display and selection 
	var <namesItems;	// array of name-item pairs for deleting by index
	var <selection;	// the currently selected item
	var <selectActions; // actions to do whenever selection changes
	*new { | list, add, remove, makeNames, fromData, toData |
		^this.newCopyArgs(nil,
			list ? #[],
			add ?? {#{ |l, i| l.list.add(i) }},
			remove ?? {#{ |l, i| l.list.remove(i); l.list }},
			makeNames ?? {#{ |l| l.list.collect { |i| [i.asString, i] }}},
			fromData ?? {#{ |d| d}},
			toData ?? {#{ |d| d}}
		).makeNamesItems;
	}
	onSelect { | ... args |
		// accepts args: view, itemAction, nilAction, or itemAction, nilAction. 
		// creates adapter that executes itemAction whenever the selection of this
		// list changes to a non nil item, and nilAction whenever the newly selected
		// item is nil. If first argument is an SCView, then the adapter is removed 
		// when that view closes.
		var view, itemAction, nilAction, adapter, oldOnClose;
		if (args[0].isKindOf(SCView)) {
			#view, itemAction, nilAction = args;
			adapter = this.makeSelectAdapter(itemAction, nilAction);
			oldOnClose = view.onClose;
			view.onClose = {
				oldOnClose.value;
				selectActions.remove(adapter);
				};
		}{
			#itemAction, nilAction = args;
			adapter = this.makeSelectAdapter(itemAction, nilAction);
		};
		selectActions = selectActions.add(adapter);
		^adapter;
	}
	makeSelectAdapter { | itemAction, nilAction |
		if (nilAction.isNil) {
			^{ | item, name, index | itemAction.(item, name, index, this) }
		};
			^{ | item, name, index |
				if (item.isNil) {
					nilAction.(item, name, index, this)
				}{
					itemAction.(item, name, index, this)
			}
		}
	}
	add { | item |
		this.list = add.(this, item);
	}
	addSelect { | item |
		this.add(item);
		this.selectItem(item);
	}
	remove { | item |
		// does not select new item in place of removed selection
		// see removeAt
		this.list = remove.(this, item);
	}
	list_ { | argList |
		list = argList;
		this.makeNamesItems;
		this.listChanged;
	}
	makeNamesItems {
		namesItems = makeNames.(this);
	}
	listChanged {
		this.changed(\list, this.indexOf(selection) ? 0, this.names);
	}
	indexOf { | item |
		// [[1, 2]].slice(nil, 1) should return [2], not 2, so
		// namesItems.slice(nil, 0) does not work consistently for [[a, b]], so use:
		if (namesItems.size == 0) {
			^nil
		}{
			^namesItems.flop[1].indexOf(item);
		}
	}
	names {
		// namesItems.slice(nil, 0) does not work consistently for [[a, b]], so use:
		if (namesItems.size == 0) {
			^[]
		}{
			^namesItems.flop[0];
		}
	}
	addListView { | listview |
		var adapter, oldOnClose;
		adapter = { | who, what, index, data |
			switch (what,
				// list changed: update items and index
				\list, {
					listview.items = data; listview.value = index ? 0; }
			)
		};
		this.addDependant(adapter);
		this.onSelect(listview, {|a,b,i|
			listview.value = i?0;
		});
		oldOnClose = listview.onClose;
		if (oldOnClose.notNil) {
			listview.onClose = { oldOnClose.value; this.removeDependant(adapter); };
		}{
			listview.onClose = { this.removeDependant(adapter); };
		};
		listview.items = this.names;
		listview.action = { |me| this.selectAt(me.value.asInteger) };
	}
	selectAt { | index |
		var itemName;
		index = index.asInteger;
		#itemName, selection = if (namesItems.size > 0) {namesItems[index]} {#["", nil]};
		this.selectionChanged(index, itemName, selection);
		^selection;
	}
	selectLast {
		if (list.size == 0) { ^nil };
		this.selectAt(list.size - 1);
	}
	selectItem { | item |
		// semantics: use selectItem as name for this method, not select, because: 
		// select is used for selecting in collections and patterns. 
		var index;
		selection = item;
		index = this.indexOf(item) ? 0;
		this.selectionChanged(index, (namesItems[index] ? [])[0], item);
	}
	selectionChanged { | index, itemName, item |
		selectActions do: { | foo | foo.(item, itemName, index); };
//		this.changed(\selection, index, itemName, item);
	}
	removeAt { | index |
		// remove at index sent by list view. Select the new
		// item at the index of a list view after removal.
		index = index.asInteger;
		this.remove(this.atIndex(index));
		this.selectAt(if (index >= list.size) { 0 } { index });
	}
	removeItem { | item |
		// remove item and select the next item in the list.
		// Must go over removeAt to get that index
		var index;
		index = list.indexOf(item);
		if (index.notNil) { this.removeAt(index); };
	}
	atIndex { | index |
		^(namesItems[index.asInteger] ? [])[1];
	}
	itemNamed { | name |
		^namesItems.detect({|i| i.first == name}).asArray.wrapAt(1);
	}
	fromData { | dataArray |
		list = fromData.(dataArray);
		this.listChanged;
	}
	toData {
		^toData.(list);
	}
	size { ^list.size }
}


NamedListModel : ListModel {

	var <>name;	// provide name variable. 
	*new { | name, list, add, remove, makeNames, fromData, toData |
		^super.new(list, add, remove, makeNames, fromData, toData)
			.name_(name);
	}
	asString { ^this.class.name.asString ++ "(" ++ name.asString ++ ")" }
}

ListWithGui : NamedListModel {
// { provide customizeable makeGui function, window and bounds variables.
	// Thus recreate a window gui for this list at the place last closed. 
	var >makeGui;
	var <window;
	var <>bounds;	// bounds of window before closing: reopen at last position
// }
	*new { | name, makeGui, list, add, remove, makeNames, fromData, toData |
		^super.new(name, list, add, remove, makeNames, fromData, toData)
			.makeGui_(makeGui);
	}
	name_ { | argName |
		name = argName;
		this.changed(\name, name);
	}
	makeGui {
		if (window.notNil) { ^window.front };
		window = makeGui.(this).onClose_({
			bounds = window.bounds;
			this.changed(\windowClosed);
			window = nil;
		}).front;
		this.listChanged;
//		this.selectItem(selection);
		this.selectAt(this.indexOf(selection) ? 0);
		^window;
	}
	closeGui {
		if (window.notNil) { window.close }
	}
	toggleGui {
		if (window.isNil) { this.makeGui } { window.close }
	}
}
