/* (IZ 2005-09-08) {
Encapsulates name, spec, action MIDIResponder, and source code for MIDIResponder function for a parameter of a script. Parameters of are created by initScript in Script. 
They are used to create control gui and to edit MIDIResponders for a script. 

Subclasses of Parameter (or even other classes) may be used to create custom guis for different types of data. The class name to be used can be provided as first element of specs in Method new. 

About MIDIResponder data: 
Source code for the MIDIResponder is needed for saving to file. Source code is built into another function code that sets the variable "script" to the script, in the shared scope of these two functions (see method: makeFunction) 
} */

Parameter : Model {
// {
	/* set by script at instance creation time: */
	var <>script;		// the script to which this parameter belongs
	var <>spec;		// spec of parameter for scaling input value
	var <>action; 	// action performed when setting this parameter
	var <name;		// name of parameter controlled. Is set by name_ 
	var <>midiName;	// for updating gui when active status changes
	/* set by activateMidi, deactivateMidi */
	var <>isActive = false;		// whether midiResponder is active
	/* set upon request, by getMidiFuncSource, getResponder */
	var <>midiFuncSource;	// for responder's function, binding script and spec vars
	var <midiResponder;			// MIDIResponder
// }
	*new { | script, specs |
		var key, myClass;
		/* { If class-name for the parameter is provided in specs[0], then create instance 
			of that named class. Otherwise try to guess the class by the parameter name provided 
			in specs[0]
		} */
		key = specs[0];		// get the key for possible class
		myClass = key.asClass;	// get the class if given
		if (myClass.isNil) {	// if no class was given, 
			myClass = switch (key,	// then guess class from parameter name: 
				\bufnum, { BufParameter }, // BufParameter
				{ this }		// default class is Parameter
			);
		}{					// else if class was actually given, then 
			specs = specs[1..];	// treat the rest of the specs as specs
		};
		^myClass.newCopyArgs(nil, script).init(specs);
	}

	init { | specs |
		var setAction;
// construct midiName, specs, action, midiFuncSource from specs
		this.name_(specs[0]);
		spec = ControlSpec(*specs[1..5]);
		action = script.makeAction(name, specs[6]);
	}

	name_ { | argName |
		name = argName;
		midiName = name ++ '_midi_';
	}
	reset { // reset scripts' parameter value to default value in spec
		action.(spec.default, spec.unmap(spec.default));
	}
	makeGui { | gui, adapterEnvir |
		var label, slider, numbox, dragsink;
		label = SCStaticText(gui, Rect(0,0,100,20)).string_(name);
		slider = SCSlider(gui, Rect(0,0,140,20));
		numbox = SCNumberBox(gui, Rect(0,0,50,20));
		dragsink = SCDragSink(gui, Rect(0,0,100,20))
			.string_("drag input here")
			.font_(Font("Helvetica", 10));
		slider.action = { | me |
			var fullVal, normalizedVal;
			fullVal = spec.map(me.value);
			normalizedVal = me.value;
			action.(fullVal, normalizedVal);
			script.changed(name, fullVal, normalizedVal);
		};
		numbox.action = { | me |
			var fullVal, normalizedVal;
			fullVal = spec.constrain(me.value);
			normalizedVal = spec.unmap(me.value);
			action.(fullVal, normalizedVal);
			script.changed(name, fullVal, normalizedVal);
		};
		numbox.value = script.envir[name];
		slider.value = spec.unmap(numbox.value);
		// update views when parameter at key changes: 
		adapterEnvir[name] = { | fullVal, unmappedVal |
			{	numbox.value = fullVal;
				slider.value = unmappedVal ?? { spec.unmap(fullVal); };
			}.defer;
		};
		// update slider knob color when midi responder activation state changes: 
		adapterEnvir[midiName] = { | isOn |
			if (isOn) {
				slider.knobColor = Color.red;
			}{
				slider.knobColor = Color.clear;
			}
		};
		dragsink.canReceiveDragHandler = {
			this.canReceiveDragHandler(dragsink, numbox, gui);
		};
		dragsink.receiveDragHandler = {
			this.receiveDragHandler(dragsink, numbox, gui);
		};
//		this.makeDragSinkActions(dragsink, numbox, gui);
	}
	activateMidi {
		// activate midi reception under the stored MIDIResponder
		if (midiResponder.isNil) {
			^nil;
		//	^Post << "NO RESPONDER! Cannot activate " << name << "\n"
		};
		if (isActive.not) {
			isActive = true;
			midiResponder.install;
			script.changed(midiName, true);
		};
	}
	deactivateMidi {
		// deactivate midi reception under the stored MIDIResponder
		if (midiResponder.isNil) {
			^nil;
		//	^Post << "NO RESPONDER! Cannot deactivate " << name << "\n"
		};
		if (isActive) {
			isActive = false;
			midiResponder.remove;
			script.changed(midiName, false);
		}
	}
	canReceiveDragHandler { | dragsink, numbox, gui |
		var object;
		object = SCView.currentDrag;
		if (object.isKindOf(MIDIResponder)) { ^true };
		^SCView.currentDrag.isKindOf(Script) and: {
			script.doesNotContainCycles(SCView.currentDrag);
		};
	}
	receiveDragHandler { | dragsink, numbox, gui |
		var object;
		object = SCView.currentDrag;
		if (object.isKindOf(MIDIResponder)) {
			^this.midiResponder_(object);
		};
		this.prReceiveDragHandler(dragsink, numbox, gui);
	}
	prReceiveDragHandler { | dragsink, numbox, gui |
		var size;
		script.addInput(dragsink.object, name, \out);
		size = script.inputs[name].writers.size;
		if (size <= 1) {
			dragsink.string = dragsink.object.name;
		}{
			dragsink.string = size.asString ++ " Scripts";
		};
	}
	getSaveData { // for saving to file
		^[midiFuncSource, isActive, midiResponder]
	}
	setSaveData { | argFCode, argIsActive, argResponder |
		midiFuncSource = argFCode;
		isActive = argIsActive;
		this.midiResponder_(argResponder);
	}
	midiResponder_ { | argResponder |
		// { store a copy of the responder you received from MIDIConfig to control yourself.
		// Provide your own function to this copy.  This function has access to yourself
		// and your spec. Make this functions based on any custom data provided by
		// script, by yourself, or construct your own default. Function source code
		// should be editable for each individual script on separate gui, but not on
		// the MIDIListener, which is only for finding out generic MIDI bindings }
		if (midiResponder.notNil) { midiResponder.remove };
		midiResponder = argResponder.copy; // make your private copy 
		midiResponder.function_(		   // and supply it with your private action func
			{ | port, chan, num, vel |
				var fullVal, normalizedVal;
				normalizedVal = vel / 127;
				fullVal = spec.map(normalizedVal);
				action.(fullVal, normalizedVal);
				script.changed(name, fullVal, normalizedVal);
			}
		);
		if (isActive) { midiResponder.install }
	}
	makeMidiResponderFunction {
/* { make function for responder from midiFuncSource.
	Provide script and spec through closure scope. 
} */
		^("{ | script, spec, action, name | " ++ this.getMidiFuncSource ++ "}"
		).interpret.(script, spec, action, name);
	}
	getMidiFuncSource {
		var fromScript;
		if (midiFuncSource.notNil) { ^midiFuncSource };
		if ((fromScript = script.envir[(name ++ '_MidiFuncSource').asSymbol]).notNil) {
			^midiFuncSource = fromScript;
		};
		^midiFuncSource = this.defaultMidiFuncSource;
	}
	defaultMidiFuncSource {
		^
"		{ | port, chan, num, vel |\
			var fullVal, normalizedVal;\
			normalizedVal = vel/127;\
			fullVal = spec.map(normalizedVal);\
			action.(fullVal, normalizedVal);\
			script.changed(name, fullVal, normalizedVal);\
		}";
	}
	canAutoMidiSetup {
		^[\in, \out].includes(name).not;
	}

/*
	close {
		// sent by Script to its parameters when it closes
		// used by some parameters to release their dependency on other objects
		// TODO: I may want to add some bus i/o releasing actions to method close
	}
*/
}

