/* (IZ 2005-10-16) { 
	UNDER CONSTRUCTION!
	Holds the configuration of a working session. 
	Saves and loads onto file this configuration. 
	The data will be: 
	- set of script groups shown on SimplePathManager instances
	- Paths of SynthDefs loaded in SynthDefs
	- Paths of buffers loaded in Samples
	??? - further configuration data like 
		buffer-arrays, 
		wavetables, 
		audio files on disk
		audio i/o channels
		servers ?

As gui, Session opens one window with:
-	a menu of all currently available session instances, 
-	a list of the script groups of the currently selected session
-	a menu with commands for opening and saving the currently selected session
 
}*/


Session {
	// {
	classvar <all;	// a ListWithGui for multiple sessions (under construction)
	// INSTANCE VARIABLES:
	var <>name;		// for displaying the session
	var <scriptGroups;	// all script groups, stored in a ListModel
					// each scriptgroup is a ListWithGui, containing a list of scripts
	var <>path;		// path for saving / loading this Session
	var <midiConfig;	// current MIDI configuration, for setting MIDI actions of scripts
	// }
	*makeGui {
		if (all.isNil) { this.initAll; };
		^all.makeGui;
	}
	*closeGui {
		if (all.notNil) { all.closeGui }
	}
	*initAll {
		// create ListWithGui to put sessions in. This is evaluated lazily, because doing it
		// at initClass involves too many dependencies on other classes being initialized.
		// *initAll is called by *makeGui.
		// TODO: Try to make this work as *initClass by resolving the initClass dependencies.
		all = ListWithGui(
			name: "Sessions",
			/* { Create window with menu for selecting a session and list view showing
				the script groups of the currently selected session.
				Keyboard action on the script group list: 
					return = show gui of selected script group 
			} */
			makeGui: { | sessions | // sessions is the ListWithGui in 'all', holding all sessions
				var window, menu, scriptGroupList, scriptGroupListView, decorator;
				var session, newSession;
				window = SCWindow("Sessions", Rect(850, 0, 150, 120).fromTop);
				this.changed(\windowOpened);
				window.view.decorator = decorator = FlowLayout(window.view.bounds, 3@3, 2@2);
				// menu with all sesssions
				menu = SCPopUpMenu(window, Rect(0,0,decorator.rest[0], 20));
				menu.keyDownAction = { | view, char, mod, uni, key |
					switch (uni,
						3, {      // enter: add new session
							newSession = this.newAdd;
							("Enter key on Session menu. Added session: " ++ newSession.name).postln;
							newSession.currentScriptGroup.makeGui;
							Session.makeGui; // bring focus back to session gui
						},
						13, {      // return: add new script group
							[this, 'return key: adding new script group'].postln;
							if ((session = sessions.atIndex(view.value)).notNil) {
								session.addScriptGroup;
								session.scriptGroups.selection.makeGui;
								Session.makeGui; // bring focus back to session gui
							};
						},
						127, {    // backspace: close + delete current session
							// TODO: rewrite this as Session-close method
							// use it also for closing from menu
							// when closing, also close all related script group guis and 
							// stop and close their scripts
							if ((session = sessions.atIndex(view.value)).notNil) {
								Confirm(
"Backspace key entered on Session Menu!\
Do you really want to remove the session named " ++ session.name ++ "?",
									{ |ok|
										if (ok) { session.remove; };
									})
								}
						},
						{ view.defaultKeyDownAction(char, mod, uni, key) }
					)
				};
				sessions.addListView(menu);
				decorator.nextLine;
				// list the script groups of the selected session:
				scriptGroupListView = SCListView(window, Rect(0, 0, *(decorator.rest - #[0, 23])));
				scriptGroupListView.keyDownAction = { | view, char, mod, uni, key |
					// coding in separate method to return if nil and for clarity
					this.performScriptGroupActions(view, char, mod, uni, key)
				};
				sessions.onSelect(menu, {|session|
					// get rid of your attachment to the previous session's script list:
					scriptGroupListView.onClose.value;
					scriptGroupListView.items_([]).value_(0);
					// attach to the script groups of the newly selected session:
					if (session.notNil) {
						session.scriptGroups.addListView(scriptGroupListView);
						// update the views index to reflect currently selected script group
						scriptGroupListView.value =
							session.scriptGroups.indexOf(session.scriptGroups.selection) ? 0;
						session.scriptGroups.onSelect(scriptGroupListView, { | script |
							script.postln;
						},{
							"no script selected".postln;
						});
						// session.postln;
					} {
						"no session selected".postln;
					}
				});
				menu.focus(true);
				decorator.nextLine;
				menu = SCPopUpMenu(window, Rect(0, 0, *decorator.rest));
				menu.items = ["(-MENU:", "Open...", "Save", "Save as...", "Close", // "Revert...", // )
					"Configure MIDI"];
				menu.action = { | me |
					[{}, {this.openDialog}, {this.save}, {this.saveAs}, {this.close},
//					{ this.revert },
					{ this.configureMidi }
					][me.value].value;
					me.value = 0;	// reselect first item
				};
				window;
			},
			list: [],
			makeNames: { |l| l.list collect: {|s| [s.name, s] }}
		);
		all.add(this.new('Default Session'));
		// make Session class update its dependants like its list with model does:.
		// Used by BrowserControlWindow script controller, and can be used elsewhere. 
		all.addDependant { | who ... whats |
			this.changed(*whats);
		};
		all.selectAt(0);
		all.selection.scriptGroups.selection.makeGui;
	}
	*doIfSelected { | action |
		var session;
		session = all.selection;
		if (session.isNil) { ^nil };
		action.(session);
	}
	*openDialog {
		"not yet implemented".postln;
	}
	*save {
		this.doIfSelected { | session | session.save; }
	}
	// TODO !!! RADICALLY REWRITE SAVE AND LOAD MECHANISMS TO SAVE SESSIONS AS SCRIPTS
	save {
		var file;
		if (path.isNil) {
			this.saveAs;
		}{
			file = File(path, "w");
			Post << "Starting to save session: '" << name << "' ... \n";
			this.saveAsScript2File(file);
			file.close;
			Post << "Saved " << name << " under: " << path;
		}
	}
	saveAsScript2File { | file |
		// save session data as script to file
		file.putAll([
			"//\tSaving session '",
			name, "' as script\n//\t",
			Date.getDate.asString,
			"\n"
		]);
		file.putAll([
			// place the start button out of the way: 
			"~script.guiBounds = Rect(0, 0, 200, 27).fromTop.fromRight;\n",
			// on start, create new session: 
			"~start = {\n\tvar session;\n",
			"session = Session.newAdd(", name.asString, "');\n"

		]);

//		file.putString("\n" ++ Array.rand(10, 0, 10).asString);
//		file.putString("\nDONE!");
	}
	*saveAs {
		this.doIfSelected { | session | session.saveAs; }
	}
	saveAs {
		CocoaDialog.savePanel({ | argPath |
			path = argPath;
			this.save;
		}, {
			"Save session cancelled".postln;
		})
	}
	getSaveData {
		^(	name: name,
			scriptGroups: scriptGroups.list.collect {|g|
				(name: g.name, bounds: g.bounds, list: g.list.collect {|s| s.saveData });
			},
			path: path,
			midiConfig: midiConfig.getSaveData
		);
	}
	*fromSaveData { | data |
		var scriptGroups;
		scriptGroups = data.scriptGroups collect: { | sg |
			
		};
		^this.new(
			data.name,
			[],
			data.path,
			MIDIConfig.fromSaveData(data.midiConfig)
		);
	}
	*close {
		this.doIfSelected { | session |
			session.closeGui;
			all.removeAt(all.indexOf(session));
		}
	}
	*performScriptGroupActions { | view, char, mod, uni, key |
		/* { perform keyboard action on the selected script group, based on 
			keyboard command entered on the listview of the script groups of the session
			in the main session window. enter: add new script group,
			space: toggle (open/close) group gui  
			return: start group, escape: stop group
			backspace: remove group. 
			Note: I am using a "hack" to determine if shift or control is pressed 
			independently of mac model (g4 pb or g5 desktop):
			if modifier value <= 256, then no modifier key is pressed.
			if 200000 > modifier > 256 then shift key is pressed
			if modifier > 200000 then the control key is pressed
		} */
		var session, groupList, group;
		if ((session = all.selection).isNil) { "no session selected".postln; ^nil };
		groupList = session.scriptGroups;
		group = groupList.selection;
		if (uni == 3) {	// enter: create new script group
				session.addScriptGroup;
				session.currentScriptGroup.makeGui;
				Session.makeGui; // bring focus back to session gui
		};
		if (group.isNil) { ^view.defaultKeyDownAction(char, mod, uni, key) };
		switch (uni,
			127, {	// backspace: close and remove script group
				group.list.do { |s|
					s.closeGui;
					s.stop;
				};
				group.closeGui;
				groupList.removeAt(view.value);
			},
			13, {	// return: start/open GUI/activate MIDI for all
				case (
					// plain return: start scripts of script group
					{ mod <= 256 }, { group.list.do { |s| s.start } },
					// shift return: show script guis of script group
					{ mod < 200000 }, {
						group.list.do { |s| s.makeGui };
						Session.makeGui; // bring focus back to session gui
					},
					// control return: activate MIDI bindings of script group
					{ true }, { group.list.do { |s| s.activateMidi } }
				);
			},
			27, { // escape: stop/close GUI/deactivate MIDI for all
				case (
					// plain escape: stop scripts
					{ mod <= 256 }, { group.list.do { |s| s.stop } },
					// shift escape: close script guis
					{ mod < 200000 }, { group.list.do { |s| s.closeGui } },
					// control escape: deactivate MIDI bindings
					{ true }, { group.list.do { |s| s.deactivateMidi } }
				);
			},
			32, { // space: toggle scriptGroup gui open/close
				group.toggleGui;
				Session.makeGui; // bring focus back to session gui
			},
			{ ^view.defaultKeyDownAction(char, mod, uni, key) }
		);
		view.focus(true);
	}
	*current {
		^all.selection
	}
	*new { | name = 'Session', scriptGroups, path, midiConfig |
		^this.newCopyArgs(
			name,
			/* { Store list of script groups in a ListModel. ListModel handles 
				display and selection of groups in the list view of the Session gui.
				The listview is attached to the gui from the Session gui whenever a new
				Session is selected. 
			} */
			ListModel(
				list: scriptGroups ? [],
				makeNames: {|l| l.list collect: {|s| [s.name, s] }}
			),
			path
		).init(midiConfig);
	}
	init { | argMidiConfig |
		if (scriptGroups.list.size == 0) { this.addScriptGroup };
		this.midiConfig =
			argMidiConfig ?? { MIDIConfig("Default MIDI Configuration", 8, 8, 8, 8) };
	}
	*newAdd { | name, scriptGroups, path, midiConfig |
		// make new session and add it to all
		var newSession;
		if (all.isNil) { this.initAll };
		newSession = this.new(
			name ?? { all.names.makeUniqueName('Session') },
			scriptGroups,
			path,
			midiConfig
		);
		all.add(newSession);
		all.selectItem(newSession);
		^newSession;
	}
	addScriptGroup { | scripts |
		scripts = scripts ? [];
		scriptGroups.addSelect(ListWithGui(
			name: scriptGroups.names.makeUniqueName('Scripts'),
			list: scripts,
			makeNames: { |me| me.list.collect {|i| [i.name, i] }},
			// gui for script groups:
			makeGui: { |listModel|
				var window, decorator, sessionlabel, listview;
				window = SCWindow(listModel.name,
					(listModel.bounds ?? {Rect(0,0,200,300).fromTop.fromRight})
				);
				window.view.decorator = decorator = FlowLayout(window.view.bounds, 3@3, 2@2);
				SCStaticText(window, Rect(0, 0, decorator.rest[0], 23))
					.align_(\center)
					.string = "(" ++ this.name.asString ++ ")";
				decorator.nextLine;
				listview = SCListView(window, Rect(0,0,*decorator.rest));
				listModel.addListView(listview);
				listview.keyDownAction = { | view, char, mod, uni, key |
					var script;
					script = listModel.selection;
					if (script.notNil) {
						switch (uni,
							127, { // backspace (delete): stop script and remove it
								script.closeGui;
								script.stop;
								listModel.removeAt(view.value);
							},
							13, {	// return
								case ( // see note on hack above for determining modifier key
									// plain return: start script
									{ mod <= 256 }, { script.start },
									// shift return: show script gui
									{ mod < 200000 }, { script.makeGui },
									// control return: activate MIDI bindings
									{ true }, { script.activateMidi }
								);
							},
							27, { // escape
								case (
									// plain escape: stop script
									{ mod <= 256 }, { script.stop },
									// shift escape: close script gui
									{ mod < 200000 }, { script.closeGui },
									// control escape: deactivate MIDI bindings
									{ true }, { script.deactivateMidi }
								);
							},
							32, { // space
								case (
									// plain space: toggle script on/off
									{ mod <= 256 }, { script.toggle },
									// shift space: toggle script gui open/close
									{ mod < 200000 }, { script.toggleGui },
									// control space: toggle activate/deactivate MIDI bindings
									{ true }, { script.toggleMidi }
								);
							},
							3, { // enter: make copy of this script to the same script group
								// TODO: IMPLEMENT copy script to script group on enter key!
							},
							{ view.defaultKeyDownAction(char, mod, uni, key) }
						)
					};
				};
				window;
			}
		).addDependant(this.makeGuiPlacerAction));
	}
	makeGuiPlacerAction { | x = 0, y = 420 |
		// { create dependant function that uses its external scope variables to 
		//	create non-overlapping gui bounds for each new script of each group }
		var minimumX, minimumY;
		minimumX = x;
		minimumY = y;
		^{ | who, what, script |
			var bounds, width, height;
			if (what == \newScript) {
				bounds = script.guiBounds;
				width = bounds.width;
				height = bounds.height;
				// if window bounds exceed right edge of screen...
				if ((x + width) > SCWindow.screenBounds.width) {
					y = y + 20; // ...then move down one row
					// ... but never lower than 100 from bottom of screen
					if ((y + 100) > SCWindow.screenBounds.height) { y = minimumY; };
					// ... also indent next row (!), but never beyond 500 
					x = minimumX = minimumX + 20 % 500;
				};
				// set the bounds of this script's gui window
				script.guiBounds = bounds.moveTo(x, y).fromTop;
				// move right of this window for next window;
				x = (x+width);
			}
		}
	}
	*currentScriptGroup {
		if (all.selection.isNil) {
			^nil
		}{
			^all.selection.currentScriptGroup;
		}
	}
	currentScriptGroup {
		^scriptGroups.selection;
	}
	*addScript { | script |
		// Add script to current group of current session. See instance methods.
		^all.selection.addScript(script);
	}
	addScript { | script |
		// { Add script to the current script group and select it
		//	If script with same name exists, append counter to new script name
		// Place the scripts guiBounds so that guis of the same groups scripts
		// do not overlap. }
		var current;
		current = scriptGroups.selection;
		script.name = current.names.makeUniqueName(script.name.asSymbol);
		current.addSelect(script).changed(\newScript, script);
		^script;
	}
	addScriptAt { | folderName, scriptName |
		// Add script found under folderName, scriptName in the ScriptBrowser 
		// to the currently selected scriptGroup of the currenltly selected Session.
		^this.addScript(Script.at(folderName, scriptName));
	}
	midiConfig_ { | argMidiConfig |
		if (midiConfig.notNil) { midiConfig.stopLearn.close.removeDependant(this); };
		midiConfig = argMidiConfig;
		midiConfig.addDependant(this);
	}
	*configureMidi {
		if(all.selection.notNil) { all.selection.configureMidi };
	}
	configureMidi {
		midiConfig.makeGui;
	}
	update { | who, what ... args |
		switch (what,
			\midiConfig, { this.midiConfig = args[0]; }
		)
	}
	close { // close all script group guis and all scripts
		scriptGroups.list.do { |sg|
			sg.closeGui;
			sg.list.do {|s| s.close }
		};
	}
	remove { // close this session and remove it from all
		this.close;
		all.removeItem(this);
	}
	bindMidi {
		// a fast and dirty method for binding all MIDIable parameters of all scripts
		var save;
		save = Script.midiConfig;			// backup
		Script.midiConfig = midiConfig;	// push yours
		this.allScripts.do { |s| s.bindMidi };
		Script.midiConfig = save;			// restore original
	}
	activateMidi {
		// activate midi bindings in all your scripts
		this.allScripts.do { |s| s.activateMidi };
	}
	deactivateMidi {
		// deactivate midi bindings in all your scripts
		this.allScripts.do { |s| s.deactivateMidi };
	}
	allScripts {
		^scriptGroups.list.collect({|sg| sg.list }).flat;
	}
	asString {
		^"Session("++name.asCompileString++")"
	}
}
