/* IZ 050723
As of 051015 rework using SimplePathManager

Creates a window containing an array of script-gui widgets. A script can be put on any of these widgets by
drag and drop from the drag-source view of the ScriptBrowser window (see class ScriptBrowser). The resulting
script patch (configuration) can be saved on file. A previously saved patch can be loaded from file into the
window of a ScriptPatch.

To make a patch portable even when the scripts it contains have moved to another folder than the paths which
they were originally loaded from, the patch saves the text of the scripts and reloads the scripts from the
script text, rather than relying on finding the original scripts at the paths where they were loaded.

As yet only the scripts are saved and loaded.

Still need to implement saving and loading of:
- connections (inputs/output) between scripts
- midi bindings
- snapshots (presets) of parameters settings.

*/

// UNDER RECONSTRUCTION
ScriptPatch : Model {
	classvar <all;	// all currently open patches;
	var <scripts; // all my scripts in an array
	var <>path;	// path where i save-reload my scripts and their configuration
	var <window;	// my gui

/*  // Utility:
ScriptPatch.loadDefault;
*/

	*initClass {
		all = Set.new;
	}

	*loadDefault {
		var path;
		path = filenameSymbol.asString.dirname.dirname ++ "/ScriptPatches/Default Script Patch";
		^this.new.makeGui.load(path);
	}

	*new {
		^super.new.init;
	}

	init {
		scripts = { Script.new; } ! 29;
		all.add(this);
	}

	makeGui {  | name = "Scripts", rect |
		if ( window.isNil) { this.makeWindow(name, rect); };
		this.makeMenu(window);
		scripts.do { |s| s.makeGui(window) };
		window.front;
	}

	makeWindow { | name = "Scripts", rect |
		window = SCWindow(name, rect ?? { Rect(350, 30, 533, 220) });
		window.view.decorator = FlowLayout(window.view.bounds, 3@3, 1@1);
		window.onClose = { window = nil };
	}

	makeMenu { | argWindow |
		var menu;
		menu = SCPopUpMenu(argWindow, Rect(0, 0, 175, 20));
		menu.items = ["( Menu: ", "Save", "Save as ...", "Open ...", "Revert"];
		menu.action = { | me |
			[nil, { this.save }, { this.saveDialog }, { this.openDialog },
				{ this.revert }
			][me.value].value;
			me.value = 0;
		};
	}

	saveDialog {
		CocoaDialog.savePanel({ | argPath |
			this.save(argPath);
		}, {
			"save cancelled".postln;
		});
	}

	save { | argPath |
		argPath = argPath ? path;
		if (argPath.isNil) { ^this.saveDialog; };
		path = argPath;
		scripts.collect { |s|
			[s.name, s.path, s.script]
		}.writeArchive(path);
		("Saved Patch: " ++ path).postln;
		this.nameWindow;
	}

	nameWindow {
		if (window.notNil) {
			window.name = (path ? "/Scripts").basename.splitext.first;
		}
	}

	openDialog {
		CocoaDialog.getPaths({ | paths |
			this.load(paths.first);
		}, {
			"load cancelled".postln;
		});
	}

	load { | argPath |
		var read;
		if (argPath.isNil) { ^this.openDialog };
		read = Object.readArchive(argPath);
		scripts.do { |s, i| s.fromSpecs(*read[i]); };
		path = argPath;
		this.nameWindow;
	}

	revert { this.load(path) }

}